# SPDX-License-Identifier: Apache-2.0

import numpy as np
from onnx import TensorProto
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import guess_numpy_type
from ..algebra.onnx_ops import (
    OnnxAdd, OnnxSub, OnnxPow, OnnxDiv, OnnxMul,
    OnnxCast, OnnxNot, OnnxLess, OnnxLog, OnnxNeg,
    OnnxImputer, OnnxIdentity, OnnxScaler)


def convert_powertransformer(scope: Scope, operator: Operator,
                             container: ModelComponentContainer):
    """Converter for PowerTransformer"""
    op_in = operator.inputs[0]
    op_out = operator.outputs[0].full_name
    op = operator.raw_operator
    opv = container.target_opset
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    lambdas = op.lambdas_.astype(dtype)

    # tensors of units and zeros
    ones_ = OnnxDiv(op_in, op_in, op_version=opv)
    zeros_ = OnnxSub(op_in, op_in, op_version=opv)

    # logical masks for input
    less_than_zero = OnnxLess(op_in, zeros_, op_version=opv)
    less_mask = OnnxCast(less_than_zero,
                         to=getattr(TensorProto, 'FLOAT'),
                         op_version=opv)

    greater_than_zero = OnnxNot(less_than_zero, op_version=opv)
    greater_mask = OnnxCast(greater_than_zero,
                            to=getattr(TensorProto, 'FLOAT'),
                            op_version=opv)

    # logical masks for lambdas
    lambda_zero_mask = np.float32(lambdas == 0)
    lambda_nonzero_mask = np.float32(lambdas != 0)
    lambda_two_mask = np.float32(lambdas == 2)
    lambda_nontwo_mask = np.float32(lambdas != 2)

    if 'yeo-johnson' in op.method:
        y0 = OnnxAdd(op_in, ones_, op_version=opv)  # For positive input
        y1 = OnnxSub(ones_, op_in, op_version=opv)  # For negative input

        # positive input, lambda != 0
        y_gr0_l_ne0 = OnnxPow(y0, lambdas, op_version=opv)
        y_gr0_l_ne0 = OnnxSub(y_gr0_l_ne0, ones_, op_version=opv)
        y_gr0_l_ne0 = OnnxDiv(y_gr0_l_ne0, lambdas, op_version=opv)
        y_gr0_l_ne0 = OnnxImputer(
            y_gr0_l_ne0, imputed_value_floats=[0.0],
            replaced_value_float=np.inf, op_version=opv)
        y_gr0_l_ne0 = OnnxMul(y_gr0_l_ne0, lambda_nonzero_mask,
                              op_version=opv)

        # positive input, lambda == 0
        y_gr0_l_eq0 = OnnxLog(y0, op_version=opv)
        y_gr0_l_eq0 = OnnxMul(y_gr0_l_eq0, lambda_zero_mask,
                              op_version=opv)

        # positive input, an arbitrary lambda
        y_gr0 = OnnxAdd(y_gr0_l_ne0, y_gr0_l_eq0, op_version=opv)
        y_gr0 = OnnxImputer(y_gr0, imputed_value_floats=[0.0],
                            replaced_value_float=np.NAN,
                            op_version=opv)
        y_gr0 = OnnxMul(y_gr0, greater_mask, op_version=opv)

        # negative input, lambda != 2
        y_le0_l_ne2 = OnnxPow(y1, 2 - lambdas, op_version=opv)
        y_le0_l_ne2 = OnnxSub(ones_, y_le0_l_ne2, op_version=opv)
        y_le0_l_ne2 = OnnxDiv(
            y_le0_l_ne2, (2 - lambdas).astype(dtype), op_version=opv)
        y_le0_l_ne2 = OnnxImputer(
            y_le0_l_ne2, imputed_value_floats=[0.0],
            replaced_value_float=np.inf, op_version=opv)
        y_le0_l_ne2 = OnnxMul(y_le0_l_ne2, lambda_nontwo_mask, op_version=opv)

        # negative input, lambda == 2
        y_le0_l_eq2 = OnnxNeg(OnnxLog(y1, op_version=opv), op_version=opv)
        y_le0_l_eq2 = OnnxMul(y_le0_l_eq2, lambda_two_mask, op_version=opv)

        # negative input, an arbitrary lambda
        y_le0 = OnnxAdd(y_le0_l_ne2, y_le0_l_eq2, op_version=opv)
        y_le0 = OnnxImputer(y_le0, imputed_value_floats=[0.0],
                            replaced_value_float=np.NAN,
                            op_version=opv)
        y_le0 = OnnxMul(y_le0, less_mask, op_version=opv)

        # Arbitrary input and lambda
        y = OnnxAdd(y_gr0, y_le0, op_version=opv)

    elif 'box-cox' in op.method:
        # positive input, lambda != 0
        y_gr0_l_ne0 = OnnxPow(op_in, lambdas, op_version=opv)
        y_gr0_l_ne0 = OnnxSub(y_gr0_l_ne0, ones_, op_version=opv)
        y_gr0_l_ne0 = OnnxDiv(y_gr0_l_ne0, lambdas, op_version=opv)
        y_gr0_l_ne0 = OnnxImputer(y_gr0_l_ne0,
                                  imputed_value_floats=[0.0],
                                  replaced_value_float=np.inf,
                                  op_version=opv)
        y_gr0_l_ne0 = OnnxMul(y_gr0_l_ne0, lambda_nonzero_mask,
                              op_version=opv)

        # positive input, lambda == 0
        y_gr0_l_eq0 = OnnxLog(op_in, op_version=opv)
        y_gr0_l_eq0 = OnnxImputer(y_gr0_l_eq0,
                                  imputed_value_floats=[0.0],
                                  replaced_value_float=np.NAN,
                                  op_version=opv)
        y_gr0_l_eq0 = OnnxMul(y_gr0_l_eq0, lambda_zero_mask, op_version=opv)

        # positive input, arbitrary lambda
        y = OnnxAdd(y_gr0_l_ne0, y_gr0_l_eq0, op_version=opv)

        # negative input
        # PowerTransformer(method='box-cox').fit(negative_data)
        # raises ValueError.
        # Therefore we cannot use convert_sklearn() for that model
    else:
        raise NotImplementedError(
            'Method {} is not supported'.format(op.method))

    y.set_onnx_name_prefix('pref')

    if op.standardize:
        use_scaler_op = container.is_allowed({'Scaler'})
        if not use_scaler_op or dtype != np.float32:
            sub = OnnxSub(y, op._scaler.mean_.astype(dtype), op_version=opv)
            final = OnnxDiv(sub, op._scaler.scale_.astype(dtype),
                            op_version=opv, output_names=[op_out])
        else:
            final = OnnxScaler(
                y, offset=op._scaler.mean_.astype(dtype),
                scale=(1.0 / op._scaler.scale_).astype(dtype),
                op_version=opv, output_names=[op_out])
    else:
        final = OnnxIdentity(y, op_version=opv, output_names=[op_out])

    final.add_to(scope, container)


register_converter('SklearnPowerTransformer', convert_powertransformer)



# SPDX-License-Identifier: Apache-2.0

"""
Main entry point to the converter from the *scikit-learn* to *onnx*.
"""
__version__ = "1.14.0"
__author__ = "Microsoft"
__producer__ = "skl2onnx"
__producer_version__ = __version__
__domain__ = "ai.onnx"
__model_version__ = 0
__max_supported_opset__ = 18  # Converters are tested up to this version.


from .convert import convert_sklearn, to_onnx, wrap_as_onnx_mixin  # noqa
from ._supported_operators import (  # noqa
    update_registered_converter, get_model_alias
)
from ._parse import update_registered_parser  # noqa
from .proto import get_latest_tested_opset_version  # noqa


def supported_converters(from_sklearn=False):
    """
    Returns the list of supported converters.
    To find the converter associated to a specific model,
    the library gets the name of the model class,
    adds ``'Sklearn'`` as a prefix and retrieves
    the associated converter if available.

    :param from_sklearn: every supported model is mapped to converter
        by a name prefixed with ``'Sklearn'``, the prefix is removed
        if this parameter is False but the function only returns converters
        whose name is prefixed by ``'Sklearn'``
    :return: list of supported models as string
    """
    from .common._registration import _converter_pool  # noqa
    # The two following lines populates the list of supported converters.
    from . import shape_calculators  # noqa
    from . import operator_converters  # noqa

    names = sorted(_converter_pool.keys())
    if from_sklearn:
        return [_[7:] for _ in names if _.startswith('Sklearn')]
    return list(names)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import (
    FloatTensorType, DoubleTensorType)
from ..common.utils import (
    check_input_and_output_numbers,
    check_input_and_output_types)


def calculate_sklearn_kernel_pca_output_shapes(operator):
    check_input_and_output_numbers(
        operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, DoubleTensorType],
        good_output_types=[FloatTensorType, DoubleTensorType])
    N = operator.inputs[0].get_first_dimension()
    op = operator.raw_operator
    lbd = op.eigenvalues_ if hasattr(op, 'eigenvalues_') else op.lambdas_
    C = lbd.shape[0]
    operator.outputs[0].type.shape = [N, C]


def calculate_sklearn_kernel_centerer_output_shapes(operator):
    check_input_and_output_numbers(
        operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, DoubleTensorType],
        good_output_types=[FloatTensorType, DoubleTensorType])
    N = operator.inputs[0].get_first_dimension()
    C = operator.raw_operator.K_fit_rows_.shape[0]
    operator.outputs[0].type.shape = [N, C]


register_shape_calculator('SklearnKernelCenterer',
                          calculate_sklearn_kernel_centerer_output_shapes)
register_shape_calculator('SklearnKernelPCA',
                          calculate_sklearn_kernel_pca_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from sklearn.base import is_regressor
from ..proto import onnx_proto
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import apply_cast, apply_concat, apply_reshape
from ..common.data_types import guess_proto_type, Int64TensorType
from .._supported_operators import sklearn_operator_name_map


def _iteration_one_versus(scope, container, inputs, i, estimator, cl_type,
                          proto_dtype, use_raw_scores=True, prob_shape=None):
    op_type = sklearn_operator_name_map[type(estimator)]

    this_operator = scope.declare_local_operator(op_type, raw_model=estimator)
    this_operator.inputs = inputs

    if is_regressor(estimator):
        score_name = scope.declare_local_variable('score_%d' % i, cl_type())
        this_operator.outputs.append(score_name)

        if hasattr(estimator, 'coef_') and len(estimator.coef_.shape) == 2:
            raise RuntimeError(
                "OneVsRestClassifier or OneVsOneClassifier accepts "
                "regressor with only one target.")
        p1 = score_name.onnx_name
        return None, None, p1

    if container.has_options(estimator, 'raw_scores'):
        options = {'raw_scores': use_raw_scores}
    elif container.has_options(estimator, 'zipmap'):
        options = {'zipmap': False}
    else:
        options = None
    if options is not None:
        container.add_options(id(estimator), options)
        scope.add_options(id(estimator), options)

    label_name = scope.declare_local_variable(
        'label_%d' % i, Int64TensorType())
    prob_name = scope.declare_local_variable(
        'proba_%d' % i, inputs[0].type.__class__())
    this_operator.outputs.append(label_name)
    this_operator.outputs.append(prob_name)

    # gets the label for the class 1
    label = scope.get_unique_variable_name('lab_%d' % i)
    apply_reshape(scope, label_name.onnx_name, label, container,
                  desired_shape=(-1, 1))
    cast_label = scope.get_unique_variable_name('cast_lab_%d' % i)
    apply_cast(scope, label, cast_label, container,
               to=proto_dtype)

    # get the probability for the class 1
    if prob_shape is None:
        # shape to use to reshape score
        cst0 = scope.get_unique_variable_name('cst0')
        container.add_initializer(cst0, onnx_proto.TensorProto.INT64, [1], [0])
        shape = scope.get_unique_variable_name('shape')
        container.add_node('Shape', [inputs[0].full_name], [shape])
        first_dim = scope.get_unique_variable_name('dim')
        container.add_node('Gather', [shape, cst0], [first_dim])
        cst_1 = scope.get_unique_variable_name('cst_1')
        container.add_initializer(
            cst_1, onnx_proto.TensorProto.INT64, [1], [-1])
        prob_shape = scope.get_unique_variable_name('shape')
        apply_concat(scope, [first_dim, cst_1], prob_shape, container, axis=0)

    prob_reshaped = scope.get_unique_variable_name('prob_%d' % i)
    container.add_node('Reshape', [prob_name.onnx_name, prob_shape],
                       [prob_reshaped])

    cst1 = scope.get_unique_variable_name('cst1')
    container.add_initializer(cst1, onnx_proto.TensorProto.INT64, [1], [1])
    cst2 = scope.get_unique_variable_name('cst2')
    container.add_initializer(cst2, onnx_proto.TensorProto.INT64, [1], [2])

    prob1 = scope.get_unique_variable_name('prob1_%d' % i)
    container.add_node(
        'Slice', [prob_reshaped, cst1, cst2, cst1], prob1)
    return prob_shape, cast_label, prob1


def convert_one_vs_one_classifier(scope: Scope, operator: Operator,
                                  container: ModelComponentContainer):

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT
    op = operator.raw_operator

    # shape to use to reshape score
    cst0 = scope.get_unique_variable_name('cst0')
    container.add_initializer(cst0, onnx_proto.TensorProto.INT64, [1], [0])
    cst1 = scope.get_unique_variable_name('cst1')
    container.add_initializer(cst1, onnx_proto.TensorProto.INT64, [1], [1])
    cst2 = scope.get_unique_variable_name('cst2')
    container.add_initializer(cst2, onnx_proto.TensorProto.INT64, [1], [2])
    shape = scope.get_unique_variable_name('shape')
    container.add_node('Shape', [operator.inputs[0].full_name], [shape])
    first_dim = scope.get_unique_variable_name('dim')
    container.add_node('Gather', [shape, cst0], [first_dim])
    cst_1 = scope.get_unique_variable_name('cst_1')
    container.add_initializer(cst_1, onnx_proto.TensorProto.INT64, [1], [-1])
    prob_shape = scope.get_unique_variable_name('shape')
    apply_concat(scope, [first_dim, cst_1], prob_shape, container, axis=0)

    label_names = []
    prob_names = []
    prob_shape = None
    cl_type = operator.inputs[0].type.__class__
    for i, estimator in enumerate(op.estimators_):
        prob_shape, cast_label, prob1 = _iteration_one_versus(
            scope, container, operator.inputs, i, estimator, cl_type,
            proto_dtype, True, prob_shape=prob_shape)

        label_names.append(cast_label)
        prob_names.append(prob1)

    conc_lab_name = scope.get_unique_variable_name('concat_out_ovo_label')
    apply_concat(scope, label_names, conc_lab_name, container, axis=1)
    conc_prob_name = scope.get_unique_variable_name('concat_out_ovo_prob')
    apply_concat(scope, prob_names, conc_prob_name, container, axis=1)

    # calls _ovr_decision_function
    this_operator = scope.declare_local_operator(
        "SklearnOVRDecisionFunction", op)

    cl_type = operator.inputs[0].type.__class__
    label = scope.declare_local_variable("label", cl_type())
    container.add_node('Identity', [conc_lab_name], [label.onnx_name])
    prob_score = scope.declare_local_variable("prob_score", cl_type())
    container.add_node('Identity', [conc_prob_name], [prob_score.onnx_name])

    this_operator.inputs.append(label)
    this_operator.inputs.append(prob_score)

    ovr_name = scope.declare_local_variable('ovr_output', cl_type())
    this_operator.outputs.append(ovr_name)

    output_name = operator.outputs[1].full_name
    container.add_node('Identity', [ovr_name.onnx_name], [output_name])

    container.add_node(
        'ArgMax', 'ovr_output', operator.outputs[0].full_name, axis=1)


register_converter('SklearnOneVsOneClassifier',
                   convert_one_vs_one_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'nocl': [True, False],
                            'output_class_labels': [False, True]})



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.utils import (
    check_input_and_output_numbers, check_input_and_output_types)
from ..common.shape_calculator import calculate_linear_regressor_output_shapes
from ..common.data_types import (
    BooleanTensorType, DoubleTensorType,
    FloatTensorType, Int64TensorType)


def calculate_bayesian_ridge_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C] ---> [N, 1]

    This operator produces a scalar prediction for every example in a
    batch. If the input batch size is N, the output shape may be
    [N, 1].
    """
    check_input_and_output_numbers(operator, input_count_range=1,
                                   output_count_range=[1, 2])
    check_input_and_output_types(operator, good_input_types=[
        BooleanTensorType, DoubleTensorType,
        FloatTensorType, Int64TensorType])

    inp0 = operator.inputs[0].type
    if isinstance(inp0, (FloatTensorType, DoubleTensorType)):
        cls_type = inp0.__class__
    else:
        cls_type = FloatTensorType

    N = operator.inputs[0].get_first_dimension()
    if (hasattr(operator.raw_operator, 'coef_') and
            len(operator.raw_operator.coef_.shape) > 1):
        operator.outputs[0].type = cls_type([
            N, operator.raw_operator.coef_.shape[1]])
    else:
        operator.outputs[0].type = cls_type([N, 1])

    if len(operator.inputs) == 2:
        # option return_std is True
        operator.outputs[1].type = cls_type([N, 1])


register_shape_calculator('SklearnAdaBoostRegressor',
                          calculate_linear_regressor_output_shapes)
register_shape_calculator('SklearnBaggingRegressor',
                          calculate_linear_regressor_output_shapes)
register_shape_calculator('SklearnBayesianRidge',
                          calculate_bayesian_ridge_output_shapes)
register_shape_calculator('SklearnLinearRegressor',
                          calculate_linear_regressor_output_shapes)
register_shape_calculator('SklearnLinearSVR',
                          calculate_linear_regressor_output_shapes)
register_shape_calculator('SklearnMLPRegressor',
                          calculate_linear_regressor_output_shapes)
register_shape_calculator('SklearnPoissonRegressor',
                          calculate_linear_regressor_output_shapes)
register_shape_calculator('SklearnRANSACRegressor',
                          calculate_linear_regressor_output_shapes)
register_shape_calculator('SklearnStackingRegressor',
                          calculate_linear_regressor_output_shapes)
register_shape_calculator('SklearnTweedieRegressor',
                          calculate_linear_regressor_output_shapes)
register_shape_calculator('SklearnGammaRegressor',
                          calculate_linear_regressor_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ..common._registration import register_shape_calculator


def pipeline_shape_calculator(operator):
    pass


def feature_union_shape_calculator(operator):
    pass


def column_transformer_shape_calculator(operator):
    pass


register_shape_calculator(
    'SklearnPipeline', pipeline_shape_calculator)
register_shape_calculator(
    'SklearnFeatureUnion', feature_union_shape_calculator)
register_shape_calculator(
    'SklearnColumnTransformer', column_transformer_shape_calculator)
